class Kaleidoscope
{
  int     m_glDrawListId;
  int     m_nb,m_nb2,m_nb2Save;  
  float   m_rot, m_rotSpeed;
  Blob    m_blob;
  boolean m_drawWired;
  videoTexture m_videoTexture;
  boolean m_is2D;

  Kaleidoscope(int nb, int nb2, Movie m)
  {
    m_nb = max(1, nb);
    m_nb2 = m_nb2Save = max(1, nb2);
    m_glDrawListId = opengl.glGenLists(1);
    m_drawWired = false;
    m_rot = 0.0f;
    m_rotSpeed = 45.0f;

    m_videoTexture = new videoTexture();
    m_videoTexture.setMovie(m);
    
    m_is2D = false;
  }

  void M_updateList(Blob b)
  {
    m_blob = b;

    BlobTriangle bTri;
    EdgeVertex eA,eB,eC;
    float xA,yA;
    float xB,yB;
    float xC,yC;
    float u = m_videoTexture.getU();
    float v = m_videoTexture.getV();
    opengl.glNewList(m_glDrawListId,GL.GL_COMPILE);
    //
    if (m_drawWired)
    {
      opengl.glBegin(GL.GL_LINES);
      for (int t=0;t<b.getTriangleNb();t++)
      {
        bTri = b.getTriangle(t);
        eA = b.getTriangleVertexA(bTri);
        eB = b.getTriangleVertexB(bTri);
        eC = b.getTriangleVertexC(bTri);

        xA = -(eA.x-m_blob.x) ; 
        yA = (eA.y-m_blob.y);  // centered
        xB = -(eB.x-m_blob.x) ; 
        yB = (eB.y-m_blob.y);
        xC = -(eC.x-m_blob.x) ; 
        yC = (eC.y-m_blob.y);

        opengl.glVertex2f(xA, yA);
        opengl.glVertex2f(xB, yB);
        opengl.glVertex2f(xB, yB);
        opengl.glVertex2f(xC, yC);
        opengl.glVertex2f(xC, yC);
        opengl.glVertex2f(xA, yA);
      }
      opengl.glEnd();
    }
    else
    {
      opengl.glBegin(GL.GL_TRIANGLES);
      for (int t=0;t<b.getTriangleNb();t++)
      {
        bTri = b.getTriangle(t);
        eA = b.getTriangleVertexA(bTri);
        eB = b.getTriangleVertexB(bTri);
        eC = b.getTriangleVertexC(bTri);

        xA = -(eA.x-m_blob.x) ; 
        yA = (eA.y-m_blob.y);  // centered
        xB = -(eB.x-m_blob.x) ; 
        yB = (eB.y-m_blob.y);
        xC = -(eC.x-m_blob.x) ; 
        yC = (eC.y-m_blob.y);


        opengl.glTexCoord2f(u*eA.x,v*eA.y);opengl.glVertex2f(xA, yA);
        opengl.glTexCoord2f(u*eB.x,v*eB.y);opengl.glVertex2f(xB, yB);
        opengl.glTexCoord2f(u*eC.x,v*eC.y);opengl.glVertex2f(xC, yC);
      }
      opengl.glEnd();
    
    }
    //
    opengl.glEndList();  

  }

  void M_wired()
  {
    m_drawWired = !m_drawWired;
    if (m_drawWired)
      smooth();
    else
      noSmooth(); 
  }
  
  void M_2D3D()
  {
    m_is2D = !m_is2D;
    if (m_is2D)
    {
      m_nb2Save = m_nb2;
      m_nb2 = 1;
    }
    else
    {
      m_nb2 = m_nb2Save;
    }
  }

  void M_update(float dt)
  {
    m_rot += m_rotSpeed*dt;
    if (m_rot >= 360.0f)
      m_rot -= 360.0f;
  }

  void M_updateVideo()
  {
    m_videoTexture.update();
  }  

  void M_draw()
  {
    if (m_blob == null)
      return;



    if (m_drawWired)
    {
      opengl.glColor4f(1.0f,0.2f,0.1f,0.3f);
    }
    else
    {
      opengl.glColor4f(1.0f,1.0f,1.0f,0.5f); // Modulating color can be changed
      opengl.glEnable(GL.GL_TEXTURE_2D);
      opengl.glBindTexture(GL.GL_TEXTURE_2D, m_videoTexture.texName[0]);
      opengl.glTexEnvf(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_MODULATE);
    }

    opengl.glEnable(GL.GL_BLEND);
    opengl.glBlendFunc(GL.GL_ONE,GL.GL_ONE);  // Additive blend
    
    //opengl.glScalef(200.0f,200.0f,200.0f);
    float step = 360.0f / (float)(m_nb);
    float alpha = 0.0f;
    float step2 = 180.0f / (float)(m_nb2);
    float beta = 0.0f;
    for (int j=0;j<m_nb2;j++)
    {
      alpha = 0.0f;
      opengl.glPushMatrix();
      opengl.glRotatef(beta, 0.0f,1.0f,0.0f);
      for (int i=0;i<m_nb;i++)
      {
        opengl.glPushMatrix();
        opengl.glScalef(200.0f,200.0f,200.0f);
        opengl.glRotatef(alpha, 0.0f,0.0f,1.0f);
        opengl.glTranslatef(0.35f, 0.0f, 0.005f*(float)i); // offset in z to avoid nasty depth buffer flickering
        opengl.glCallList(m_glDrawListId); 
        opengl.glPopMatrix();
        alpha += step;
      }
      opengl.glPopMatrix();
      beta += step2;
    }

    opengl.glDisable(GL.GL_BLEND);
    if (!m_drawWired)
      opengl.glDisable(GL.GL_TEXTURE_2D);
  }
}
