//------------------------------------------------------------
// Quat class
//------------------------------------------------------------

class Quaternion
{

  float r,x,y,z;
  
 
  
  Quaternion()
  {
    setIdentity();
  }
  
  Quaternion(float r, float x, float y, float z)
  {
    set(r,x,y,z);
  }
  
  void setIdentity()
  {
    this.r = 1.0;
    this.x = 0.0;
    this.y = 0.0;
    this.z = 0.0;
  }
  
  void set(float r,float x,float y,float z)
  {
    this.r = r;
    this.x = x;
    this.y = y;
    this.z = z;  
  }
 
  void set(float r, vector3 v)
  {
      this.r = r;
      this.x = v.x;
      this.y = v.y;
      this.z = v.z;
  }

  void set(Quaternion q)
  {
      this.r = q.r;
      this.x = q.x;
      this.y = q.y;
      this.z = q.z;
  
  
  }

  void normalize()
  {
    float norm;
    norm = sqrt(r*r + x*x + y*y + z*z);

    this.r /= norm;
    this.x /= norm;
    this.y /= norm;
    this.z /= norm;
  }  
  
  boolean isUnit()
  {
    return (r*r + x*x + y*y + z*z > 0.99);
  }  
  
 void conjuguate()
  {
    this.x = -this.x;
    this.y = -this.y;
    this.z = -this.z;  
  }

 void multiplyBy(Quaternion q2)
{
	Quaternion qTemp=new Quaternion();
	qTemp.r = this.r;
	qTemp.x = this.x;
	qTemp.y = this.y;
	qTemp.z = this.z;

	this.r = qTemp.r * q2.r - (qTemp.x * q2.x + qTemp.y * q2.y + qTemp.z * q2.z);
	this.x = qTemp.r * q2.x +  q2.r * qTemp.x + qTemp.y * q2.z - qTemp.z * q2.y; 
	this.y = qTemp.r * q2.y -  q2.z * qTemp.x + qTemp.y * q2.r + qTemp.z * q2.x;
	this.z = qTemp.r * q2.z +  q2.y * qTemp.x - qTemp.y * q2.x + qTemp.z * q2.r;

}

void multiplyBy(float a)
{
	this.r *= a;
	this.x *= a;
	this.y *= a;
	this.z *= a;
}


void vectorTransform(vector3 v)
{
   this.r = 0;
   this.x = v.x;
   this.y = v.y;
   this.z = v.z;

}


float[] toMatrix3()
{
	float rx, ry, rz, xx, yy, yz, xy, xz, zz, x2, y2, z2; 
        float[] matrix = new float[9];
        
	x2 = this.x + this.x; 
	y2 = this.y + this.y; 
	z2 = this.z + this.z;
	
	xx = this.x * x2; xy = this.x * y2; xz = this.x * z2;
	yy = this.y * y2; yz = this.y * z2; zz = this.z * z2;
	rx = this.r * x2; ry = this.r * y2; rz = this.r * z2;


	matrix[0] = 1.0 - (yy + zz); 
	matrix[1] = xy + rz; 
	matrix[2] = xz - ry; 

	matrix[3] = xy - rz;
	matrix[4] = 1.0 - (xx + zz);
	matrix[5] = yz + rx;

	matrix[6] = xz + ry; 
	matrix[7] = yz - rx; 
	matrix[8] = 1.0 - (xx + yy); 

        return matrix;

}

float[] toMatrix4()
{
        float[] matrix = new float[16];
        float[] matrix3 = this.toMatrix3();
        
        matrix[0] = matrix3[0]; matrix[1] = matrix3[1] ; matrix[2] = matrix3[2]  ; matrix[3] = 0.0f ; 
        matrix[4] = matrix3[3]; matrix[5] = matrix3[4] ; matrix[6] = matrix3[5]  ; matrix[7] = 0.0f ; 
        matrix[8] = matrix3[6]; matrix[9] = matrix3[7] ; matrix[10] = matrix3[8]  ; matrix[11] = 0.0f ; 
        matrix[12] = 0.0f; matrix[13] = 0.0f ; matrix[14] = 0.0f  ; matrix[15] = 1.0f ; 
                        
        return matrix;

}

}





// -------------------------------------------------------------
// Multiply()
// q = q1*q2
// -------------------------------------------------------------

Quaternion multiply(Quaternion q1, Quaternion q2)
  {
  
    Quaternion result = new Quaternion();
    result.r = q1.r * q2.r - (q1.x * q2.x + q1.y * q2.y + q1.z * q2.z);
    result.x = q1.r * q2.x +  q2.r * q1.x + q1.y * q2.z - q1.z * q2.y; 
    result.y = q1.r * q2.y -  q2.z * q1.x + q1.y * q2.r + q1.z * q2.x;
    result.z = q1.r * q2.z +  q2.y * q1.x - q1.y * q2.x + q1.z * q2.r;
    
    return result;
  }


Quaternion multiply(Quaternion q, float a)
{
  Quaternion result = new Quaternion();
  result.r = q.r * a;
  result.x = q.x * a;
  result.y = q.y * a;
  result.z = q.z * a;
  return result;
}



// -------------------------------------------------------------
// vectorRotate()
// Rotates a vector3 , rotation given by a Quat
// result = q.v.q^-1
// -------------------------------------------------------------

vector3 vectorRotate(vector3 v, Quaternion rot)
{
      vector3 result   = new vector3();
      Quaternion invr  = new Quaternion();
      Quaternion qv    = new Quaternion();
      Quaternion tmp   = new Quaternion();
	
      invr.r =  rot.r;
      invr.x = -rot.x;
      invr.z = -rot.y;
      invr.z = -rot.z;
      
      qv.vectorTransform(v);
	
      tmp = multiply(qv, invr);
      invr = multiply(rot, tmp); 
	
	result.x = invr.x;
	result.y = invr.y;
	result.z = invr.z;	

      return result;
}

// -------------------------------------------------------------
// eulerTransform
// Takes 3 angles, and returns a Quat
// -------------------------------------------------------------

Quaternion eulerTransform(float alpha, float beta, float gamma)
{
        Quaternion result = new Quaternion();
	float cr, cp, cy, sr, sp, sy, cpcy, spsy;

	alpha *= DEG_TO_RAD;
	beta  *= DEG_TO_RAD;
	gamma *= DEG_TO_RAD;

	cr = cos(alpha/2);

	cp = cos(beta/2);
	cy = cos(gamma/2);


	sr = sin(alpha/2);
	sp = sin(beta/2);
	sy = sin(gamma/2);
	
	cpcy = cp * cy;
	spsy = sp * sy;


	result.r = cr * cpcy + sr * spsy;
	result.x = sr * cpcy - cr * spsy;
	result.y = cr * sp * cy + sr * cp * sy;
	result.z = cr * cp * sy - sr * sp * cy;

        return result;
}


// -------------------------------------------------------------
// SLERP: Spherical Linear Interpolation
// Step from q1 to q2, 0=<t=<1
// SLERP(q1,q2,0) = q1
// SLERP(q1,q2,1) = q2
// -------------------------------------------------------------

Quaternion SLERP(Quaternion q1, Quaternion q2, float t)
{
        Quaternion result = new Quaternion();
        float[]     to1 = new float[4];
        float       omega, cos_omega, sin_omega, scale0, scale1;


        // calc cosine
        cos_omega = q1.r*q2.r + q1.x*q2.x + q1.y*q2.y + q1.z*q2.z;


        // adjust signs (if necessary)
        if ( cos_omega <0.0 ){ 
        	cos_omega = -cos_omega; 
        	to1[0] = - q2.r;
        	to1[1] = - q2.x;
		to1[2] = - q2.y;
		to1[3] = - q2.z;
		
        } else  {
        	to1[0] =   q2.r;
        	to1[1] =   q2.x;
		to1[2] =   q2.y;
		to1[3] =   q2.z;
        }


        // calculate coefficients


       if ( (1.0 - cos_omega) > 0.01 ) {
                // standard case (slerp)
                omega = (float)Math.acos(cos_omega);
                sin_omega = sin(omega);
                scale0 = sin((1.0 - t) * omega) / sin_omega;
                scale1 = sin(t * omega) / sin_omega;


        } else {        
    	    // "from" and "to" Quats are very close 
	    //  ... so we can do a linear interpolation
                scale0 = 1.0 - t;
                scale1 = t;
        }
	// calculate final values
	result.r = scale0 * q1.r + scale1 * to1[0];
	result.x = scale0 * q1.x + scale1 * to1[1];
	result.y = scale0 * q1.y + scale1 * to1[2];
	result.z = scale0 * q1.z + scale1 * to1[3];

        return result;
}
