//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//Kaleidoscope by v3ga <http://www.v3ga.net>
//December 2005
//Processing 0098 Beta
//
// ~~~~~~~~~~
// software :
// ~~~~~~~~~~
// - Video library
// - BlobDetection library
//
// ~~~~~~~~~~
// keys :
// ~~~~~~~~~~
// - p : play/pause the movie
// - w : draw wired polygons
// - s : switch 2D / 3D for polygon drawing
// - a funny glitch : try pausing the video in poylgon mode, then switch to wired. The render will draw polygon with no texture + lines ! 
//
// ~~~~~~~~~~
// mouse :
// ~~~~~~~~~~
// - left button drag : rotate
// - right button drag : back and forth
//
// Seems there are some issues with depth + alpha blending (order of drawing)
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
import blobDetection.*;
import processing.video.*; 
import processing.opengl.*;
import net.java.games.jogl.*;

// Constants
String movieName = "meDark4.mov";
float th = 0.20f;


// BlobDetection
BlobDetection theBlobDetection;
Blob theBiggestBlob;
PImage imgSmall;

// Movie
Movie myMovie; 
boolean isNewVideoFrame;
boolean isMoviePaused;

// Kaleidoscope
Kaleidoscope myKaleido;

// OpenGL
GL opengl; 
GLU glu;

// Timer
timer t;
float dt;

// ArcBall
ArcBall  theArcBall;
Quaternion qTarget;
Quaternion qRotation;
float[]     mRotation;

//Camera
float zCam = 300.0f;
float zCamTarget = zCam;
float zCamStart = zCam;
float yMouseStart = 0.0f;


//RenderToTexture
RenderToTexture rtt;

// Filter
FilterRGB fxRGB;

// Background
Texture bgTexture;

// --------------------------------------------
// Application Export
// --------------------------------------------
/*
static public void main(String args[]) 
{
  PApplet.main(new String[] { "kaleidoscope_fx" });
}
*/
// --------------------------------------------
// setup()
// --------------------------------------------
void setup() 
{ 
  size(800, 600, OPENGL); 
  
  // GL init
  setupGL();

  // Movie
  myMovie = new Movie(this, movieName); 
  myMovie.loop(); 
  isNewVideoFrame = false;
  isMoviePaused = false;

  // BlobDetection
  BlobDetection.setConstants(1000,1000,1000); // Be sure to have enough room for polygon (1000 triangles max here for a blob)
  imgSmall = new PImage(40,30); // image which will be sent to detection (a smaller copy of the video frame);
  theBlobDetection = new BlobDetection(imgSmall.width, imgSmall.height);
  theBlobDetection.setPosDiscrimination(true);
  theBlobDetection.computeTriangles();
  theBlobDetection.setThreshold(th); // will detect bright areas whose luminosity > th;  

  // Kaleidoscope
  myKaleido = new Kaleidoscope(6, 5, myMovie);
  
  // RenderToTexture
  rtt = new RenderToTexture(512,512);
  
  // Filter
  fxRGB = new FilterRGB(64,7.5f,1.0f);
  
  // Background
  bgTexture = new Texture();
  bgTexture.createFromImage("bg.jpg");

  // Timer
  t = null;
  
  // Arcball
  initArcBall();
} 

// --------------------------------------------
// draw()
// --------------------------------------------
void draw() 
{
  // Timing 
  if (t==null)
    t = new timer(millis());
  dt = t.M_getDelta(millis())*0.001;

  // Video update
  if (isNewVideoFrame)
  {
    isNewVideoFrame=false;
    imgSmall.copy(myMovie, 0, 0, myMovie.width, myMovie.height, 0, 0, imgSmall.width, imgSmall.height);
    //fastblur(imgSmall, 1); 
    theBlobDetection.computeBlobs(imgSmall.pixels);
    theBiggestBlob = findBiggestBlob();

    // Update video texture
    myKaleido.M_updateVideo();
    // we refresh the gl drawing list
    myKaleido.M_updateList(theBiggestBlob);
  }

  // Update vars
  myKaleido.M_update(dt);
  fxRGB.M_update(dt);
  updateParams(dt);

     
  // Set our own matrix stuff now
  rtt.M_begin();
  setGLProjectionMatrixIdentity();
  opengl.glFrustum(-1.0,1.0,-1.0,1.0,2.0f,1000.0f);
  opengl.glClear( GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
 
  
  // Modelview
  opengl.glMatrixMode(GL.GL_MODELVIEW);
  opengl.glLoadIdentity();
  glu.gluLookAt(0.0f,0.0f,-zCam, 0.0f,0.0f,0.0f, 0.0f,1.0f,0.0f);
  
  // Kaleido
  opengl.glMultMatrixf(mRotation);
  myKaleido.M_draw();

  rtt.M_end();

 // Fullscreen
 opengl.glViewport(0,0,width, height);
 opengl.glScissor(0,0,width, height);
 setGLProjectionMatrixIdentity();
 opengl.glMatrixMode(GL.GL_MODELVIEW);
 opengl.glLoadIdentity();
 opengl.glClear( GL.GL_COLOR_BUFFER_BIT );
  
 opengl.glColor3f(1.0f,1.0f,1.0f); 
 rtt.M_bindTexture();
 fxRGB.M_apply();
 rtt.M_unbindTexture();
} 


// --------------------------------------------
// movieEvent()
// Called every time a new frame is 
// available to read 
// --------------------------------------------
void movieEvent(Movie m) 
{ 
  myMovie.read(); 
  isNewVideoFrame = true;
} 

// --------------------------------------------
// newBlobDetectedEvent()
// Filtering blobs here (discard "little" ones)
// --------------------------------------------
boolean newBlobDetectedEvent(Blob b)
{
  int w = (int)(b.w * imgSmall.width);
  int h = (int)(b.h * imgSmall.height);
  if (w >= 20 || h >= 20)
    return true;
  return false;
} 

// --------------------------------------------
// findBiggestBlob()
// --------------------------------------------
Blob findBiggestBlob()
{
  Blob biggestBlob = null;
  float surface = 0.0f;
  float surfaceMax = 0.0f;
  Blob b=null;
  for (int i=0;i<theBlobDetection.getBlobNb();i++)
  {
    b = theBlobDetection.getBlob(i);
    surface = b.w * b.h;
    if (surface > surfaceMax)
    {
      surfaceMax=surface;
      biggestBlob = b;
    }
  }

  return biggestBlob; 
}

// --------------------------------------------
// setupGL()
// --------------------------------------------
void setupGL()
{
  glu = ((PGraphicsGL) g).glu;
  opengl = ((PGraphicsGL) g).gl;
  opengl.glClearColor(0.0f,0.0f,0.0f,1.0f);
}

// --------------------------------------------
// setGLProjectionMatrix()
// --------------------------------------------
void setGLProjectionMatrixIdentity()
{
  opengl.glMatrixMode(GL.GL_PROJECTION);
  opengl.glLoadIdentity();
}

// --------------------------------------------
// mousePressed()
// --------------------------------------------
void mousePressed()
{
  if (mouseButton == LEFT)
    theArcBall.start(qTarget, (float)mouseX, (float)mouseY);
  if (mouseButton == RIGHT)
  {
    zCamStart = zCam;
    yMouseStart = mouseY;
  }
}

// --------------------------------------------
// mouseDragged()
// --------------------------------------------
void mouseDragged()
{
  if (mouseButton == LEFT)
    qTarget = theArcBall.update((float)mouseX, (float)mouseY);
  if (mouseButton == RIGHT)
    zCamTarget = zCamStart + ((float)mouseY-yMouseStart)/(float)height * 400.0f;
}

// --------------------------------------------
// keyPressed()
// --------------------------------------------
void keyPressed()
{
  if (key=='w')
    myKaleido.M_wired();
  if (key=='p')
  {
    isMoviePaused = !isMoviePaused;
    if (isMoviePaused)
      myMovie.pause();
    else
      myMovie.play();
    
  }
  if (key=='s')
    myKaleido.M_2D3D();
}

// --------------------------------------------
// initArcBall()
// --------------------------------------------
void initArcBall()
{
  theArcBall   = new ArcBall(width, height);
  qRotation = new Quaternion();
  qTarget = new Quaternion();
  qRotation.setIdentity();
  qTarget.setIdentity();

  mRotation = new float[16];
  for (int i=0;i<16;i++) mRotation[i] = 0.0f;
  mRotation[0] = 1.0f;mRotation[5] = 1.0f;mRotation[10] = 1.0f;mRotation[15] = 1.0f;

  
}

// --------------------------------------------
// updateParams()
// --------------------------------------------
void updateParams(float dt)
{
  qRotation = SLERP(qRotation, qTarget, 0.1f);
  mRotation = qRotation.toMatrix4();

  zCam += 0.1f*(zCamTarget-zCam);
  zCam = max(zCam, 100.0f);
  zCam = min(zCam, 600.0f);
}

